#!/bin/env python
"""
This program runs the FragSeq algorithm for producing cutting scores from
mapped reads.

See the README.txt file (part of this source code distribution) for usage,
contact info, how to cite, etc.

Copyright 2010 Andrew V. Uzilov and The Regents of the University of California
All Rights Reserved

This file is part of FragSeq_v0.0.1.

FragSeq_v0.0.1 is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FragSeq_v0.0.1 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FragSeq_v0.0.1.  If not, see <http://www.gnu.org/licenses/>.
"""

# Python std lib.
import array
import copy
import gzip
import itertools
import logging
import math
import optparse
import re
import sys
import time

# FragSeq libs.
import config
import wiggle


#===============================================================================
# Logging.
#===============================================================================

# Set up logger.
logging.basicConfig (format=
    '%(levelname)-8s [readsToStruct.py] %(message)s')
logger = logging.getLogger()
logger.setLevel (logging.DEBUG)  # default logging level


#===============================================================================
# Module globals.
#===============================================================================

# Valid configuration attributes.
# See documentation for libavu.config module for how this is interpreted.
validConfigAttrs = {
    'infile': {
        'coords':  None,
        'stock': None,
        'reads': {
            'ssbg': None, 
            'ss':   None,
            },
        },
    'outfile': {
        'reads': {
            'ss':   None,
            'ssbg': None,
            'local': int,
            },
        'rawends': {
            'ss':   None,
            'ssbg': None,
            },
        'cutscores': {
            'ss':   None,
            },
        'figure': {
            'noSizes':  None,
            'sizesNuc': None,
            'sizesBg':  None,
            },
        'stats':    None,  
        'config':   None,
        'endDist': {
            'ss':   None,
            'ssbg': None,
            'trimOnly': int,
            },
        'wiggle': {
            'raw': {
                'ss': {
                    'plus':  None,
                    'minus': None,
                    },
                'ssbg': {
                    'plus':  None,
                    'minus': None,
                    },
                },

            'probs': {
                'ss': {
                    'plus':  None,
                    'minus': None,
                    },
                'ssbg': {
                    'plus':  None,
                    'minus': None,
                    },
                },

            'ss': {
                'plus':  None,
                'minus': None,
                },
            },
        },
    'wiggle': {
        'raw': {
            'ss': {
                'plus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                'minus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                },
            'ssbg': {
                'plus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                'minus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                },
            },

        'probs': {
            'ss': {
                'plus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                'minus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                },
            'ssbg': {
                'plus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                'minus': {
                    'trackLabel':  None,
                    'description': None,
                    'visibility':  None,
                    'color':       None,
                    'extra':       None,
                    },
                },
            },

        'ss': {
            'plus': {
                'trackLabel':  None,
                'description': None,
                'visibility':  None,
                'color':       None,
                'extra':       None,
                },
            'minus': {
                'trackLabel':  None,
                'description': None,
                'visibility':  None,
                'color':       None,
                'extra':       None,
                },
            },
        },
    'logScores': int,
    'countMethod': ['3no', '3all', '3trim', '3lost'],
    'noMultClvg': int,
    'trimLenMax': int,
    'sizeRange': {
        'lower': int,
        'upper': int,
        },
    'pseudocounts': {
        'ss':   float,
        'ssbg': float,
        'auto': int,
        },
    'ignoreNoDataSites': int,
    'noiseCutoff': int,
    'posOnly': int,
    'numEndSitesToIgnore': int,
    }


# IUPAC table as a dict.
# Key: IUPAC code, value: set of valid nucs.
# Only uppercase, RNA nucs are used.
iupacTable = {
    'A': set ('A'),
    'C': set ('C'),
    'G': set ('G'),
    'U': set ('U'),
    'R': set ('AG'),
    'Y': set ('CU'),
    'M': set ('AC'),
    'K': set ('GU'),
    'S': set ('CG'),
    'W': set ('AU'),
    'B': set ('CGU'),
    'D': set ('AGU'),
    'H': set ('ACU'),
    'V': set ('ACG'),
    'N': set ('ACGU')}
validNucs = set (iupacTable.keys())


#===============================================================================
# Classes and factory functions.
#===============================================================================

class Feature():
    """Simple little class for representing a genomic location.
    """
    def __init__ (self, name, landmark, L, R, strand,
                  multMapCount=None, seq=None, score=None):
        # Safety.
        L, R = int (L), int (R)
        assert L <= R
        assert strand == '+' or strand == '-'
        if multMapCount is not None:
            multMapCount = int (multMapCount)
        if seq is not None:
            assert len (seq) == R - L
        if score is not None:
            score = float (score)
        
        # Save.
        self.name         = name
        self.landmark     = landmark
        self.L            = L
        self.R            = R
        self.strand       = strand
        self.multMapCount = multMapCount
        self.seq          = seq
        self.score        = score
    
    def __len__ (self):
        return self.R - self.L


#===============================================================================
# Module-level helper functions.
#===============================================================================

def cmpNucSeqs (seqA, seqB, useIupac=True):
    """Compares two equal-length nucleic acid sequences for
    position-by-position similarity, returning a list of booleans that
    specifies "match/no match" for every position.
    
    The returned list of booleans is therefore the same length as the input
    seqs.

    Comparison is case insensitive and treats T and U as equal (even when
    "useIupac" is off), so DNA can be compared to RNA and so forth.
    
    If "useIupac" is on, the comparison accounts for IUPAC degenerate codes,
    favoring maximal matching.  That is, if IUPAC degenerate codes are used,
    a position is considered a match if the degenerate sets have at least one
    nucleotide in common.  So, all of the following will match each other
    because they all have an adenine in common: A, R, M, W, D, H, V, N.
    Note that N therefore matches every nucleotide.
    
    If "useIupac" is off, a strict comparison is performed.  The characters
    must match exactly, but aren't checked for biological validity.
    Note that N will therefore only match another N.
    """
    assert len (seqA) == len (seqB), \
           "Input sequences must be of equal length, but you passed in:\n%s (len=%d)\n%s (len=%d)" \
           % (seqA, len (seqA), seqB, len (seqB))
    
    # Put nucs in a standard form.
    seqA = seqA.upper().replace ('T', 'U')  # RNA is more fun
    seqB = seqB.upper().replace ('T', 'U')

    # Validate input.
    for i in xrange (len (seqA)):
        if seqA[i] not in validNucs:
            assert False, \
                   'Invalid nuc char "%s" in seqA at pos %d (0-based).' \
                   % (seqA[i], i)
        if seqB[i] not in validNucs:
            assert False, \
                   'Invalid nuc char "%s" in seqB at pos %d (0-based).' \
                   % (seqB[i], i)

    # Do position-by-position comparison.
    ret = [True for x in seqA]

    if useIupac:
        for i in xrange (len (seqA)):
            setA = iupacTable[seqA[i]]
            setB = iupacTable[seqB[i]]
            if setA.isdisjoint (setB):
                # No nucleotides in common can exist, even allowing for IUPAC
                # degeneracy.
                ret[i] = False
    else:
        for i in xrange (len (seqA)):
            if seqA[i] != seqB[i]:
                ret[i] = False
    
    return ret


def loadBedAsDict (bedFileName):
    '''Load first six columns of a BED file and return as a dictionary mapping
    feature name (string) to feature (Feature object).
    
    Names in BED file must be unique.
    '''
    logger.debug ('Loading BED (as dict) from file "%s"...' % bedFileName)

    if bedFileName[-3:] == '.gz':
        bedFile = gzip.GzipFile (filename=bedFileName, mode='rb')
    else:
        bedFile = open (bedFileName, 'r')

    ret = dict()
    for line in bedFile:
        if line.startswith ('track') or line.startswith ('browser'):
            continue
        
        fields = line.strip().split()
        if len (fields) < 6:
            logger.debug ('Skipping weird BED line in file "%s":\n%s'
                          % (bedFileName, line))
            continue
        landmark, L, R, name, score, strand = fields[0:6]
        # Don't allow duplicates.
        if ret.has_key (name):
            logger.critical ('Name "%s" in BED input file "%s" not unique.'
                             % (name, bedFileName))
            sys.exit (1)
        ret[name] = Feature (name, landmark, L, R, strand)
    return ret


def loadBedAsIter (bedFileNames, saveName=False, saveSeq=False,
                   saveMultMapCount=False, saveScore=False):
    """Returns an iterator over BED file(s) that lazily (i.e. efficiently in
    memory) loads each line and returns it as a Feature object.

    Input is either a single BED file name, or a non-empty list of BED file
    names.  Iterator will iterate over all features in all BED files, in the
    order they go in the list.

    If 'saveSeq' and/or 'saveMultMapCount' and/or 'saveName' are set to True,
    then every BED feature name will be assumed to be something like:
    
        1000_1039_73_F3:CCATTGCACTTTGGGGTGTGCTGACCCCTGCGAgTT-1
    
    and the colorspace read name, inferred sequence, and multiple mapping
    suffix will be extracted and saved into the 'name', 'seq', and/or
    'multMapCount' Feature object properties, respectively.

    The above also applies for 'saveScore', except that the score is taken
    from the BED data score column (col 5) and saved to the 'score' property
    """
    if not isinstance (bedFileNames, list):
        # Loading from a single BED file.
        logger.debug ('Loading BED (as iterator) from a file named: %s'
                      % bedFileNames)
        bedFileNames = [bedFileNames]
    else:
        # Loading from several BED files.
        assert len (bedFileNames) >= 1
        logger.debug ('Loading BED (as iterator) from %d file(s) named: %s'
                      % (len (bedFileNames), ' '.join (bedFileNames)))
    
    parseName = re.compile (r'(\w+):(\w+)[-+](\d+)$')
    
    for bedFileName in bedFileNames:
        if bedFileName[-3:] == '.gz':
            bedFile = gzip.GzipFile (filename=bedFileName, mode='rb')
        else:
            bedFile = open (bedFileName, 'r')

        for line in bedFile:
            if line.startswith ('track') or line.startswith ('browser'):
                continue
            
            fields = line.strip().split()
            if len (fields) < 6:
                logger.debug ('Skipping weird BED line in file "%s":\n%s'
                              % (bedFileName, line))
                continue
            landmark, L, R, name, score, strand = fields[0:6]
            
            retFeat = Feature (
                None, landmark, L, R, strand, multMapCount=None, seq=None,
                score=None)
            
            if saveScore:
                retFeat.score = float (score)
            
            if saveName or saveSeq or saveMultMapCount:
                matchName = parseName.search (name.strip())
                assert matchName is not None, 'Bad name: "%s"' % name.strip()
                if saveName:
                    retFeat.name = matchName.group (1)
                if saveSeq:
                    retFeat.seq = matchName.group (2)
                if saveMultMapCount:
                    retFeat.multMapCount = int (matchName.group (3))
                    assert retFeat.multMapCount >= 1

            yield retFeat
    

def loadStk (fileName):
    """A generator function for iteration over multiple sequence alignments
    in a Stockholm-format file, where each iteration yields the 2-tuple
    (seqs, annots) containing sequences and #=GC (per-stanza, per-column)
    annotations, respectively.

    The input Stockholm file can contain several multiple sequence alignment
    stanzas (delimited by "//" in the usual Stockholm convention).  Each
    iteration yields contents of a single stanza.

    Each entry in the yielded tuple is a dictionary.  In 'seqs', the key is
    the sequence ID and the value is the sequence.  In 'annots', the key is
    the #=GC tag (also called "feature name") and the value is the annotation.
    
    Neither the sequence nor the annotation are altered or checked for
    validity, except that all sequences and annotations within a stanza must
    be of the same length.  All remaining checks to make sure that data
    makes sense are up to the caller.  This means any conventions you are using
    (e.g. uppercase/lowercase nucleotides in the sequence) are retained.
    """
    logger.debug (
        'Loading sequence(s) and structure(s) from Stockholm file "%s"...'
        % fileName)

    # Init.
    retSeqs     = {}
    retGcAnnots = {}
    
    getSeq     = re.compile (r'^([^#\s]+)\s+(\S+)\s*$')
    getGcAnnot = re.compile (r'^#=GC\s+(\S+)\s+(\S+)\s*$')
    endOfMsa   = re.compile (r'^//\s*$')

    for line in open (fileName, 'r'):
        if endOfMsa.search (line):
            # End of multiple sequence alignment stanza; yield its contents.

            # Safety.
            stanzaLen = len (retSeqs.values()[0])  # get arbitrary len
            for tag in retGcAnnots.keys():
                assert stanzaLen == len (retGcAnnots[tag])
            for seqId in retSeqs.keys():
                assert stanzaLen == len (retSeqs[seqId])

            yield (retSeqs, retGcAnnots)
                
            # Reset for next MSA stanza.
            retSeqs     = {}
            retGcAnnots = {}
            
        else:
            matchSeq = getSeq.search (line)
            if matchSeq:
                seqId = matchSeq.group (1)
                if not retSeqs.has_key (seqId):
                    retSeqs[seqId] = ''
                retSeqs[seqId] += matchSeq.group (2)
                continue

            matchGcAnnot = getGcAnnot.search (line)
            if matchGcAnnot:
                tag = matchGcAnnot.group (1)
                if not retGcAnnots.has_key (tag):
                    retGcAnnots[tag] = ''
                retGcAnnots[tag] += matchGcAnnot.group (2)
                continue


def loadReads (featBedFileName, readBedFileNames, binSize=5000):
    """Find and return reads for each feature in 'featBedFileName'.

    'readBedFileNames' is a list of BED files to scan for reads.
    File names can be None, in which case they will be skipped and a None
    object gets returned in its place...
    [note also each "name" can actually be a list of names, so you can load
    reads of same "type" from multiple files...]

    Returns a list of tuples: (Feature obj, 'reads') in 1-to-1 correspondence...

    'reads' above (and below in the code, consistently) is a list of
    lists, same in length as 'readBedFileNames',
    mapping 1-to-1 in meaning (that is, each entry is a list of reads from
    that file that are in the feature)

    Read coordinates get converted to be local to sequence.

    Note the distinction between 'features' and 'reads' below...

    ...features MUST have unique names in the BED file...
    
    For optimization, assumption is made that read number is gigantic (so they
    are lazily loaded to save memory), but feature number is manageable (so
    they are completely loaded and binned for faster access).
    """
    logger.debug ("Starting loadReads()...")
    assert isinstance (readBedFileNames, list)
    assert len (readBedFileNames) > 0
    startTime = time.time()

    features = loadBedAsDict (featBedFileName)  # dict: name -> Feature obj
    featureNames = [n for n in features.keys()]
    # Sort names.
    # This IS IMPORTANT: downstream steps may rely on Python-ordering of
    # feature names (e.g. when features are entire chromosomes for whole-genome
    # analysis).
    featureNames.sort()

    # A list of iterators through BED files, 1-to-1 with 'readCoordsFileNames';
    # each iterator loops through the respective BED file, returning Feature
    # objects.
    readIters = [
        loadBedAsIter (f, saveName=True, saveSeq=True, saveMultMapCount=True,
                       saveScore=True)
        if f is not None else None for f in readBedFileNames]
    
    logger.info ("Looking through %d groups of BED files for reads for %d sequences..."
                 % (len (readIters), len (featureNames)))

    #! For each feature, find reads that are from that feature.
    #! Read objects are copied (originals remain unchanged) and each copy's
    #! coordinates are changed to be relative to the read's feature.

    # featToReads key:   Feature name.
    # featToReads value: A list in 1-to-1 correspondence with the input
    #                    BED file list.  Each list entry is a list of reads
    #                    from that BED file that are subsequences of the
    #                    feature.
    # Summary:
    #     featToReads[featName] = [[bed1read1, bed1read2, ...],
    #                              [bed2read1, bed2read2, ...],
    #                              ...]
    featToReads = dict()
    for name in featureNames:
        # Init to empty lists of reads for each file.
        # Use None to denote 'no BED file' placeholder.
        featToReads[name] = [[] if i is not None else None for i in readIters]

    # Bin the features by strand, chromosome, and location to speed up
    # overlap searches later.  The data structure is used like this:
    #
    #     featsBinned[strand][chr][binStart] = [feat1, feat2, feat3, ...]
    #
    # where 'binStart' is the start coord of the bin.  A bin starting at
    # 'binStart' covers coordinates x that are in the range:
    #
    #     binStart <= x < binStart+binSize
    #
    # in 0-based INCLUSIVE (not half-open) coordinates.
    #
    # Note that a feature can span several bins, so a dumb concatenation of
    # several bins can create a list with duplicate features.

    featsBinned = {'+': {}, '-': {}}
    for name in featureNames:
        feat = features[name]

        if len (feat) == 0:
            # Stupid edge case that will probably never happen.
            # But if it does, it breaks later math, so let's be
            # technically correct and handle it.
            # Create an empty list of reads inside this feature
            # (there can be no read in a 0-length feature) and
            # skip handling it later.
            for i in xrange (len (readIters)):
                featToReads[feat.name][i] = []
            continue
        
        # What are the left endpoints of bin(s) containing the feature?
        # (Need to convert to 0-based inclusive coords to do math.)
        binL = feat.L - (feat.L % binSize)
        binR = (feat.R-1) - ((feat.R-1) % binSize)
        assert binL <= binR, \
               "featL=%d featR=%d; binL=%d binR=%d" % (feat.L, feat.R,
                                                       binL, binR)
        
        #if binL == binR:
        #    logger.debug (
        #        'Feature [%d,%d) is in one bin: [%d,%d)' % (
        #            feat.L, feat.R, binL, binL+binSize))
        #else:
        #    logger.debug (
        #        'Feature [%d,%d) is in several bins: %s' % (
        #            feat.L,
        #            feat.R,
        #            ', '.join (['[%d,%d)' % (b,b+binSize) for b in xrange (
        #                binL,binR+1,binSize)])
        #            ))
        
        if not featsBinned[feat.strand].has_key (feat.landmark):
            featsBinned[feat.strand][feat.landmark] = {}

        for binStart in xrange (binL, binR+1, binSize):
            if not featsBinned[feat.strand][feat.landmark].has_key (
                    binStart):
                featsBinned[feat.strand][feat.landmark][binStart] = []
            featsBinned[feat.strand][feat.landmark][binStart].append (feat)
    
    #logger.debug ('Plus strand keys: %s' % featsBinned['+'].keys())  #d!!!
    #logger.debug ('Minus strand keys: %s' % featsBinned['-'].keys())
    
    for i in xrange (len (readIters)):
        readIter = readIters[i]
        if readIter is None:
            # Skip 'no file' BED files.
            logger.debug ('Skipping loading reads from BED file %d.' % i)
            continue
        
        # Go through all reads in a BED file,
        readCount, readStepDebug = 0, 500000  # for optimization measures
        prevTime = time.time()
        for read in readIter:
            # For each read, go through each feature on same strand,
            # chromosome, and same bin(s), and see if read is from
            # the feature.

            # Time some stuff before we begin anything (can't put this at
            # end of loop because won't reach it due to "continue" statements.
            if (readCount % readStepDebug == 0) and (readCount > 0):
                logger.debug (
                    "Took %.2f seconds to process previous %d reads from %d BED file groups for %d features."
                    % (time.time() - prevTime, readStepDebug,
                       len (readIters), len (features)))
                logger.debug ("Done with %d reads." % readCount)
                prevTime = time.time()
            
            readCount += 1
            
            if not featsBinned[read.strand].has_key (read.landmark):
                # Rare edge case.
                # This read is not from a landmark for which we have features
                # on this strand; skip it.
                continue
            
            # What are the left endpoints of bin(s) containing the read?
            # (Need to convert to 0-based inclusive coords to do math.)
            binL = read.L - (read.L % binSize)
            binR = (read.R-1) - ((read.R-1) % binSize)
            assert binL <= binR, \
                "readL=%d readR=%d; binL=%d binR=%d" % (
                    read.L, read.R, binL, binR)
            
            # Get an iterable of unique features that are in the same bin(s)
            # as this read.
            if binL == binR:
                # Easy case: read is contained solely in a single bin.
                if featsBinned[read.strand][read.landmark].has_key (binL):
                    featsInSameBin = \
                        featsBinned[read.strand][read.landmark][binL]
                else:
                    # There are features in this bin, therefore read can't
                    # overlap anything, so skip it.
                    continue
            else:
                # Trickier case: read spans several bins.
                # This means that if any features in the read's bins also span
                # several bins also, they will be counted for overlap multiple
                # times.
                # We have to create a data structure where each feature is
                # represented only once.
                # This is a bit inefficient, but because this case is rare
                # (reads are small, bins are large), the occasional efficiency
                # hit is worth it for the simplicity of implementation.
                featNamesInSameBin = set()
                for binStart in xrange (binL, binR+1, binSize):
                    if featsBinned[read.strand][read.landmark].has_key (
                            binStart):
                        for f in featsBinned[read.strand][read.landmark][ \
                                binStart]:
                            featNamesInSameBin.add (f.name)
                featsInSameBin = [features[n] for n in featNamesInSameBin]
            
            for feat in featsInSameBin:
                # Is this read COMPLETELY inside this feature?
                if read.L >= feat.L and read.R <= feat.R:
                    # This read is a subinterval of the feature.
                    # Convert read coords to be relative/local to feature.
                    # This logic (and the copy op) should be in libavu
                    # (this is the "reverse" to mapToParentLandmark())
                    readLocal = copy.copy (read)
                    readLocal.landmark = feat.name
                    if feat.strand == '+':
                        readLocal.L = read.L - feat.L
                        readLocal.R = read.R - feat.L
                    else:
                        assert feat.strand == '-'
                        readLocal.L = feat.R - read.R
                        readLocal.R = feat.R - read.L
                        assert readLocal.L < readLocal.R
                        readLocal.strand = '+'
                    # Save.
                    featToReads[feat.name][i].append (readLocal)

    # Return.
    for name in featureNames:
        numReadsStr = ','.join ([str (len (l)) for l in featToReads[name]])
        logger.info ('Loaded %s reads for feature "%s".' %
                     (numReadsStr, name))
    logger.debug (
        "Done with loadReads(), took %.2f seconds, returning..."
        % (time.time() - startTime))
    return [(features[name], featToReads[name]) for name in featureNames]


def countCleavageEvents (method5p, method3p, reads, feat,
                         trimSize=None, sizeRange=None, separate=False):
    """Given a feature and its sequencing reads, infer number of cleavage
    events between each pair of nucleotides using 5' and 3' end-counting
    methods specified.

    ...describe 'method5p' and 'method3p'...

    Note that native 5' end and 3' end are also counted and termed 'cleavage'
    in this context, even though they are native ends.  That is, we count reads
    ending at 5' end and 3' end of the feature as 'cleavage' at the native ends.

    Some counting methods require "trimSize" and/or "maxSize" to be set.

    ...describe read coord space and other options...

    ...describe 'separate' option...
    
    Returns array(s) (4-byte unsigned long ints) of inferred cleavage event
    counts using 0-based in-between coordinates local to feature to index into
    the array.
    Index i describes site between base i-1 and i (using 0-based base numbering).
    Index 0 is start of feature (5' of 1st nucleotide).
    Index len (feat) is end of seq (3' of last nucleotide).

    If 'separate' is True or if only 5' or only 3' ends are counted, a single
    list is returned.  Otherwise a tuple of 2 lists is returned (for cleavage
    resulting from 5' and 3' end counts, in that order).
    """
    featLen = len (feat)
    logger.debug (
        "Counting cleavage events from %d reads for feature of len %d..."
        % (len (reads), featLen))
    
    # Init to 0.
    # Use float rather than int because we may be over-writing contents
    # in-place with floats when doing bg correction.
    logger.debug ("Allocating memory for arrays...")
    cleavageCount5p = array.array ('f', itertools.repeat (0.0, featLen + 1))
    cleavageCount3p = array.array ('f', itertools.repeat (0.0, featLen + 1))
    logger.debug ("                            ...done!")

    # Count cleavage events based on 5' ends.
    if method5p == 'yes':
        # Count only 5' ends of reads.
        for read in reads:
            cleavageCount5p[read.L] += 1
    elif method5p == 'no':
        assert method3p != 'no', \
               "You can't tell this function to count neither 5' nor 3' ends; there is no point."
    else:
        assert False, 'Invalid 5prime end counting method: "%s"' % method5p

    # Count cleavage events based on 3' ends.
    if method3p == '3all':
        # Use all 3' ends indiscriminately.
        # Results in double-counting of sites because some 3' ends come from
        # sites that also produced 5' ends that we may have already counted.
        for read in reads:
            cleavageCount3p[read.R] += 1
    elif method3p == '3trim':
        # Only count "trim" 3' ends, so that our odds of getting true 3' ends
        # are greater.
        # Still results in double-counting of sites for same reason as '3all'
        # option, but to a lesser extent.
        assert trimSize is not None
        for read in reads:
            if read.R - read.L <= trimSize:
                cleavageCount3p[read.R] += 1
    elif method3p == '3lost':
        # Only count "trim" 3' ends if the other fragment that results
        # under the assumption of single-hit kinetics is out of size
        # selection bounds, meaning we will not observe it and its 5'
        # end, therefore counting the 3' end of observed fragment is not
        # double-counting.
        assert trimSize is not None
        assert sizeRange is not None
        minSize, maxSize = sizeRange
        for read in reads:
            if read.R - read.L <= trimSize \
                   and (featLen - read.R >= maxSize
                        or featLen - read.R <= minSize):
                cleavageCount3p[read.R] += 1
    else:
        assert method3p == '3no', \
               'Invalid 3prime end counting method: "%s"' % method3p

    # Figure out what to return.
    if method5p == 'yes' and method3p == '3no':
        # Return only counts based on 5' end.
        assert not separate
        logger.debug ("Counted %d 5' ends (without pseudocounts)." 
                      % sum (cleavageCount5p))
        return cleavageCount5p
    elif method5p == 'no' and method3p != '3no':
        # Return only counts based on 3' end.
        assert not separate
        logger.debug ("Counted %d 3' ends (without pseudocounts)." 
                      % sum (cleavageCount3p))
        return cleavageCount3p
    elif separate:
        # Both ends used in counting; return 5' and 3' counts separately.
        logger.debug ("Counted %d 5' ends and %d 3' ends (without pseudocounts)."
                      % (sum (cleavageCount5p), sum (cleavageCount3p)))
        return (cleavageCount5p, cleavageCount3p)
    else:
        # Both ends used in counting; merge 5' and 3' counts.
        logger.debug ("Counted %d cleavage events (without pseudocounts)."
                      % (sum (cleavageCount5p) + sum (cleavageCount3p)))
        # Add into existing array, for memory efficiency.
        cleavageCountTotal = cleavageCount5p
        for i in xrange (featLen + 1):
            cleavageCountTotal[i] += cleavageCount3p[i]
        return cleavageCountTotal


def correctForBg (cleavageExp, cleavageBg, useLog=True):
    """Correct cleavageExp cut rates/scores for background.

    Both cleavage arrays MUST be of type 'array' filled with type 'float'.

    Zero-valued data is not allowed EXCEPT WHEN both nuclease and background
    are 0, which can only occur when 'ignoreNoDataSites' option is on, in
    which case those nucs are ignored.
    If that option is off, there should be no zero-valued data because
    pseudocounts should have been previously been added by caller.
    
    Background correction is done IN PLACE for memory efficiency.
    That is, 'cleavageExp' *IS* modified and will contain the
    background-corrected values.
    'cleavageBg' is not modified.
    """
    logger.debug ("Correcting for background...")
    assert len (cleavageExp) == len (cleavageBg)
    dataLen = len (cleavageExp)

    #! Ratio (log ratio or plain ratio).
    if useLog:
        for i in xrange (dataLen):
            if cleavageExp[i] == 0.0 or cleavageBg[i] == 0.0:
                # This should only occur if 'ignoreNoDataSites' option
                # is on, in which case values in both samples must be equal
                # to zero (no pseudocounts added to missing data nucs).
                # Do nothing but safety.
                assert cleavageExp[i] == 0.0
                assert cleavageBg[i]  == 0.0
            else:
                # Use natural log (base e).
                cleavageExp[i] = math.log (cleavageExp[i] / cleavageBg[i])
    else:
        for i in xrange (dataLen):
            if cleavageExp[i] == 0.0 or cleavageBg[i] == 0.0:
                # This should only occur if 'ignoreNoDataSites' option
                # is on, in which case values in both samples must be equal
                # to zero (no pseudocounts added to missing data nucs).
                # Do nothing but safety.
                assert cleavageExp[i] == 0.0
                assert cleavageBg[i]  == 0.0
            else:
                cleavageExp[i] /= cleavageBg[i]
    
   

#===============================================================================
# Running from the command line.
#===============================================================================

usageMessage = """
    %prog [options] configFile1.conf [configFile2.conf [...]]

You can specify one or more config files, whose contents will be merged.

To get info about options, run:
    %prog -h"""

if __name__ == '__main__':
    parser = optparse.OptionParser (usage=usageMessage)
    opts, args = parser.parse_args()

    if len (args) < 1:
        parser.error ('Need to specify one or more config file(s) as positional arguments.')

    # Load config.
    conf = config.Config (
        validConfigAttrs, configFileNameOrList=args, multValues='append',
        mergeConfigFiles=True)

    if not conf.pseudocounts.auto:
        # If pseudocounts are not defined, give them a default.
        # This ensures that downstream code doesn't break due to math on
        # NoneType objects.
        # The default must be non-zero, to ensure we have no zero-divide errors
        # in normalization downstream.
        if conf.pseudocounts.ss   is None: conf.pseudocounts.ss   = 1
        if conf.pseudocounts.ssbg is None: conf.pseudocounts.ssbg = 1
    else:
        # Resolve conflict
        assert conf.pseudocounts.ss   is None
        assert conf.pseudocounts.ssbg is None
    
    if conf.infile.stock is not None:
        # Load sequences and structures from Stockholm file.

        # All of these map unique seq ID --> value.
        seqs           = {}
        structs        = {}
        mismatchAnnots = {}

        for currSeqs, currAnnots in loadStk (conf.infile.stock):
            for seqId in currSeqs.keys():
                if seqs.has_key (seqId) or structs.has_key (seqId) \
                       or mismatchAnnots.has_key (seqId):
                    logger.critical (
                        'Sequence ID "%s" in Stockholm input file "%s" not unique.'
                        % (seqId, conf.infile.stock))
                    sys.exit (1)
                else:
                    seqs[seqId]    = currSeqs[seqId]
                    structs[seqId] = currAnnots['SS_cons']
                    # This annotation is optional.
                    if currAnnots.has_key ('checkMismatch'):
                        mismatchAnnots[seqId] = currAnnots['checkMismatch']
        
        # Standardize the secondary structure annotations.
        for seqId in structs.keys():
            structs[seqId] = rnadraw.standardizeSecStruct (structs[seqId])

        # Convert mismatch masks to more useful type.
        for seqId in mismatchAnnots.keys():
            mismatchAnnots[seqId] = [int (x) for x in mismatchAnnots[seqId]]
    
    if conf.noiseCutoff is None:
        # A sensible default; any non-zero cleavage is considered not noise.
        conf.noiseCutoff = 1

    # NB: numEndSitesToIgnore includes the native 5' and 3' ends.
    # So if you set it to 1, only the native 5' and 3' ends are ignored.
    if conf.numEndSitesToIgnore is not None:
        assert conf.numEndSitesToIgnore >= 0
    else:
        conf.numEndSitesToIgnore = 0  # default is don't ignore any end sites
    
    doRawWiggle = False
    if conf.outfile.wiggle.raw.ss.plus is not None:
        # Write wiggle track headers for raw cut count wiggle data.

        assert conf.outfile.wiggle.raw.ssbg.plus  is not None
        assert conf.outfile.wiggle.raw.ss.minus   is not None
        assert conf.outfile.wiggle.raw.ssbg.minus is not None
        doRawWiggle = True

        wiggleOutRawSsPlus = wiggle.WiggleWriter (
            conf.outfile.wiggle.raw.ss.plus,
            name            = conf.wiggle.raw.ss.plus.trackLabel,
            description     = conf.wiggle.raw.ss.plus.description,
            visibility      = conf.wiggle.raw.ss.plus.visibility,
            color           = conf.wiggle.raw.ss.plus.color,
            extra           = conf.wiggle.raw.ss.plus.extra,
            maxHeightPixels = (200, 50, 10))
        wiggleOutRawSsBgPlus = wiggle.WiggleWriter (
            conf.outfile.wiggle.raw.ssbg.plus,
            name            = conf.wiggle.raw.ssbg.plus.trackLabel,
            description     = conf.wiggle.raw.ssbg.plus.description,
            visibility      = conf.wiggle.raw.ssbg.plus.visibility,
            color           = conf.wiggle.raw.ssbg.plus.color,
            extra           = conf.wiggle.raw.ssbg.plus.extra,
            maxHeightPixels = (200, 50, 10))
        
        wiggleOutRawSsMinus = wiggle.WiggleWriter (
            conf.outfile.wiggle.raw.ss.minus,
            name            = conf.wiggle.raw.ss.minus.trackLabel,
            description     = conf.wiggle.raw.ss.minus.description,
            visibility      = conf.wiggle.raw.ss.minus.visibility,
            color           = conf.wiggle.raw.ss.minus.color,
            extra           = conf.wiggle.raw.ss.minus.extra,
            maxHeightPixels = (200, 50, 10))
        wiggleOutRawSsBgMinus = wiggle.WiggleWriter (
            conf.outfile.wiggle.raw.ssbg.minus,
            name            = conf.wiggle.raw.ssbg.minus.trackLabel,
            description     = conf.wiggle.raw.ssbg.minus.description,
            visibility      = conf.wiggle.raw.ssbg.minus.visibility,
            color           = conf.wiggle.raw.ssbg.minus.color,
            extra           = conf.wiggle.raw.ssbg.minus.extra,
            maxHeightPixels = (200, 50, 10))
        
    doProbWiggle = False
    if conf.outfile.wiggle.probs.ss.plus is not None:
        # Write wiggle track headers for break probability wiggle data.

        assert conf.outfile.wiggle.probs.ssbg.plus  is not None
        assert conf.outfile.wiggle.probs.ss.minus   is not None
        assert conf.outfile.wiggle.probs.ssbg.minus is not None
        doProbWiggle = True

        wiggleOutProbSsPlus = wiggle.WiggleWriter (
            conf.outfile.wiggle.probs.ss.plus,
            name            = conf.wiggle.probs.ss.plus.trackLabel,
            description     = conf.wiggle.probs.ss.plus.description,
            visibility      = conf.wiggle.probs.ss.plus.visibility,
            color           = conf.wiggle.probs.ss.plus.color,
            extra           = conf.wiggle.probs.ss.plus.extra,
            maxHeightPixels = (200, 50, 10))
        wiggleOutProbSsBgPlus = wiggle.WiggleWriter (
            conf.outfile.wiggle.probs.ssbg.plus,
            name            = conf.wiggle.probs.ssbg.plus.trackLabel,
            description     = conf.wiggle.probs.ssbg.plus.description,
            visibility      = conf.wiggle.probs.ssbg.plus.visibility,
            color           = conf.wiggle.probs.ssbg.plus.color,
            extra           = conf.wiggle.probs.ssbg.plus.extra,
            maxHeightPixels = (200, 50, 10))
        
        wiggleOutProbSsMinus = wiggle.WiggleWriter (
            conf.outfile.wiggle.probs.ss.minus,
            name            = conf.wiggle.probs.ss.minus.trackLabel,
            description     = conf.wiggle.probs.ss.minus.description,
            visibility      = conf.wiggle.probs.ss.minus.visibility,
            color           = conf.wiggle.probs.ss.minus.color,
            extra           = conf.wiggle.probs.ss.minus.extra,
            maxHeightPixels = (200, 50, 10))
        wiggleOutProbSsBgMinus = wiggle.WiggleWriter (
            conf.outfile.wiggle.probs.ssbg.minus,
            name            = conf.wiggle.probs.ssbg.minus.trackLabel,
            description     = conf.wiggle.probs.ssbg.minus.description,
            visibility      = conf.wiggle.probs.ssbg.minus.visibility,
            color           = conf.wiggle.probs.ssbg.minus.color,
            extra           = conf.wiggle.probs.ssbg.minus.extra,
            maxHeightPixels = (200, 50, 10))
        
    doCutScoreWiggle = False
    if conf.outfile.wiggle.ss.plus is not None:
        # Write wiggle track headers for cutting score wiggle data.
        
        assert conf.outfile.wiggle.ss.minus is not None
        doCutScoreWiggle = True

        wiggleOutCutScoreSsPlus = wiggle.WiggleWriter (
            conf.outfile.wiggle.ss.plus,
            name            = conf.wiggle.ss.plus.trackLabel,
            description     = conf.wiggle.ss.plus.description,
            visibility      = conf.wiggle.ss.plus.visibility,
            color           = conf.wiggle.ss.plus.color,
            extra           = conf.wiggle.ss.plus.extra,
            maxHeightPixels = (200, 50, 10))
        wiggleOutCutScoreSsMinus = wiggle.WiggleWriter (
            conf.outfile.wiggle.ss.minus,
            name            = conf.wiggle.ss.minus.trackLabel,
            description     = conf.wiggle.ss.minus.description,
            visibility      = conf.wiggle.ss.minus.visibility,
            color           = conf.wiggle.ss.minus.color,
            extra           = conf.wiggle.ss.minus.extra,
            maxHeightPixels = (200, 50, 10))

    #! Process the reads.

    # Save config log for this analysis run.
    # Do this HERE, right before starting the loop over features, because some
    # user-specified config settings may have gotten overwritten above.
    open (conf.outfile.config, 'w').write (str (conf) + '\n')

    # Loop over each RNA and its reads and structure (if any).
    for feat, reads in loadReads (
            conf.infile.coords,
            [conf.infile.reads.ss, conf.infile.reads.ssbg]):
        logger.debug ('Processing feature "%s"...' % feat.name)

        numSites = len (feat) + 1  # num sites b/ween bases

        # If this assert fails, we are ignoring every site in the RNA, which
        # does not make sense.
        assert numSites > conf.numEndSitesToIgnore * 2

        # Write some basic statistics about reads for this feature.
        if conf.outfile.stats is not None:
            statsFile = open (
                conf.outfile.stats.replace ('%', feat.name), 'w')
        else:
            statsFile = open ('/dev/null', 'w')  # TODO: Linux-specific hack -- fix!!
        statsFile.write (
            'Num raw trim reads loaded (before any filtering/normalization/etc): ss %d, ssbg %d\n' %
            (len (reads[0]), len (reads[1])))
        statsFile.write (
            'Coverage by raw trim reads (before any filtering/normalization/etc) in trim: ss %d, ssbg %d\n' %
            (sum ([len (r) for r in reads[0]]) / float (len (feat)),
             sum ([len (r) for r in reads[1]]) / float (len (feat))))

        # Write multiple mapping index (MMI) to stats file.
        for readsInSample, sampleName \
                in zip (reads[0:2], ('ss', 'ssbg')):
            l = len (feat) + 1
            coveragePerPos = [0 for i in xrange (l)]
            multMapPerPos  = [0 for i in xrange (l)]
            for r in readsInSample:
                for i in xrange (r.L, r.R+1):
                    coveragePerPos[i] += 1
                    multMapPerPos[i]  += r.multMapCount
            #statsFile.write ('covPerPos: %s\nmmiPerPos: %s\n' % (str (coveragePerPos), str (multMapPerPos)))  #D!!!
            mmi = 0.0
            for i in xrange (l):
                # Compute MMI per site.
                if coveragePerPos[i] != 0:
                    mmi += float (multMapPerPos[i]) \
                           / float (coveragePerPos[i])
            # Normalize MMI by RNA length, but only to count of sites
            # that had nonzero coverage.
            effectiveLen = len ([x for x in coveragePerPos if x > 0])
            if effectiveLen > 0:
                mmi /= effectiveLen
            else:
                # This happens when a RNA has zero coverage in a sample.
                mmi = 0.0
            statsFile.write ('MMI (based on raw trim reads, before any filtering/normalization/etc) for sample %s: %f\n' % (sampleName, mmi))

        # Figure out mismatches per site across all mapped reads.
        # This can only be done if we loaded the known sequence from the
        # Stockholm file.
        # Also, if a mismatch annotation line was given in the Stockholm file,
        # throw out reads with mismatches.
        if conf.infile.stock is not None:
            for i in xrange (2):
                sampleName     = ('ss', 'ssbg')[i]
                readsInSample  = reads[i]
                readsToKeep    = []
                readsToDiscard = []
                mismatchCount  = [0 for x in xrange (len (feat))]
            
                for read in readsInSample:
                    # Subseq of known seq that overlaps the read.
                    knownSeq = seqs[feat.name][read.L:read.R]
                    assert len (knownSeq) == len (read.seq)

                    matches = cmpNucSeqs (read.seq, knownSeq, useIupac=True)
                    for j, m in zip (xrange (read.L, read.R), matches):
                        if not m:
                            mismatchCount[j] += 1
                
                    # d!!!
                    #print >> sys.stderr, "Known seq: %s [%d-%d]" \
                    #      % (knownSeq, read.L+1, read.R)
                    #print >> sys.stderr, "Read seq:  %s" % read.seq
                    #print >> sys.stderr, "Mism:      %s" % ''.join ([
                    #    '.' if m else 'x' for m in matches])
                
                    keepRead = True
                    if mismatchAnnots.has_key (feat.name):
                        # Use the mismatch annotation line from the Stockholm file
                        # to see if there is a mismatch between this read and the
                        # known seq at any of the disallowed sites; if so, mark
                        # the read for throwing out.
                    
                        currMismAnnot = mismatchAnnots[feat.name][read.L:read.R]
                        assert len (matches) == len (currMismAnnot)
                    
                        for m1, m2 in zip (currMismAnnot, matches):
                            if m1 > 0 and not m2:
                                # Mismatch at this position disallowed, but a
                                # mismatch exists; so, we DON'T want this read.
                                keepRead = False
                                break
                    
                        #print >> sys.stderr, "Ban mism:  %s" % ''.join ([
                        #    str (x) for x in currMismAnnot])  #d!!!
                
                    if keepRead:
                        readsToKeep.append (read)
                        #print >> sys.stderr, "Read kept."  #d!!!
                    else:
                        readsToDiscard.append (read)
                        #print >> sys.stderr, "Read discarded."  #d!!!

                    # d!!!
                    #print >> sys.stderr, "Mismatches up to:"
                    #print >> sys.stderr, ', '.join ([
                    #    '%s%d:%d' % (seqs[feat.name][k], k+1, mismatchCount[k])
                    #    for k in xrange (len (mismatchCount))])
            
                # Overwrite original reads in this sample with filtered reads.
                reads[i] = readsToKeep

                # Log some fun info.
                statsFile.write (
                    "Read mismatches per site (based on raw trim reads, before any filtering/normalization/etc) for sample %s: %s\n"
                    % (sampleName, ', '.join ([
                        '%s%d:%d' % (seqs[feat.name][k], k+1, mismatchCount[k])
                        for k in xrange (len (mismatchCount))])))
                statsFile.write (
                    "In sample %s: kept %d reads, discarded %d reads due to mismatches.\n"
                    % (sampleName, len (readsToKeep), len (readsToDiscard)))
        
        else:
            print >> statsFile, \
                "Did not filter for mismatches because no Stockholm file given."

        # Print some data to get an idea of whether single-hit kinetics
        # assumptions hold.
        for readsInSample, sampleName \
                in zip (reads[0:2], ('ss', 'ssbg')):
            endDistFileName = getattr (conf.outfile.endDist, sampleName)
            if endDistFileName is not None:
                endDistFile = open (
                    endDistFileName.replace ('%', feat.name), 'w')
                print >> endDistFile, 'distFrom5p\tdistFrom3p'
                featR = len (feat)  # end of feat, in coords local to feat
                if conf.outfile.endDist.trimOnly:
                    # Only print ends for trim reads, since both ends are
                    # real.
                    for r in readsInSample:
                        if len (r) <= conf.trimLenMax:
                            print >> endDistFile, '%d\t%d' % (r.L, featR - r.R)
                else:
                    # Print ends for all reads indiscriminately.
                    for r in readsInSample:
                        print >> endDistFile, '%d\t%d' % (r.L, featR - r.R)
        
        if conf.noMultClvg:
            # Remove reads that were produced by more than one cleavage.
            # That is, keep reads that contain original 5' or 3' end.
            # We define 'containing original end' as starting or ending
            # within 'numEndSitesToIgnore' of the feature's start and end.
            # This loose definition is required because in many cases, the
            # exact bounds of feature are hazy.
            
            before = (len (reads[0]), len (reads[1]))
            
            featR = len (feat)  # end of feat, in coords local to feat
            for i in xrange (len (reads)):
                reads[i] = [
                    r for r in reads[i]
                    if r.L         <= conf.numEndSitesToIgnore or
                       featR - r.R <= conf.numEndSitesToIgnore]
        
            print >> statsFile, \
                "Removed %d (out of %d) reads from SS due to noMultClvg rule" \
                % (before[0] - len (reads[0]), before[0])
            print >> statsFile, \
                "Removed %d (out of %d) reads from SSBG due to noMultClvg rule" \
                % (before[1] - len (reads[1]), before[1])
        
        # Save the (possibly filtered) reads used for the analysis.
        for readsInSample, outFileName \
                in zip (reads[0:2],
                        (conf.outfile.reads.ss,   conf.outfile.reads.ssbg)):
            if outFileName is not None:
                outFile = open (outFileName.replace ('%', feat.name), 'w')
                for read in readsInSample:
                    # Reassemble read name from the pieces into which we broke
                    # it up earlier, and output in BED format.
                    # The only difference is that "+/-" convention before the
                    # multiple mapping suffix is not followed - we just use "-"
                    # regardless of the count.
                    if conf.outfile.reads.local:
                        # Keep coords of reads local to parent RNA.
                        print >> outFile, '\t'.join ([str (x) for x in [
                            read.landmark, read.L, read.R,
                            '%s:%s-%d' % (
                                read.name, read.seq, read.multMapCount),
                            read.score, read.strand]])
                    else:
                        # Convert read coords back to global.
                        readGlobal = copy.copy (read)
                        readGlobal.landmark = feat.landmark
                        if feat.strand == '+':
                            readGlobal.L += feat.L
                            readGlobal.R += feat.L
                        else:
                            assert feat.strand == '-'
                            newL = feat.R - readGlobal.R
                            newR = feat.R - readGlobal.L
                            readGlobal.L, readGlobal.R = newL, newR
                            assert readGlobal.L <= readGlobal.R
                            if readGlobal.strand == '-':
                                readGlobal.strand = '+'
                            else:
                                readGlobal.strand = '-'
                        print >> outFile, '\t'.join ([str (x) for x in [
                            readGlobal.landmark, readGlobal.L, readGlobal.R,
                            '%s:%s-%d' % (
                                readGlobal.name, readGlobal.seq,
                                readGlobal.multMapCount),
                            readGlobal.score, readGlobal.strand]])
        
        # Count cleavage events in each sample.
        logger.debug ('Counting cleavage events...')
        ssCleavage   = countCleavageEvents (
            'yes', conf.countMethod, reads[0], feat,
            trimSize=conf.trimLenMax,
            sizeRange=(conf.sizeRange.lower, conf.sizeRange.upper),
            separate=False)
        ssBgCleavage = countCleavageEvents (
            'yes', conf.countMethod, reads[1], feat,
            trimSize=conf.trimLenMax,
            sizeRange=(conf.sizeRange.lower, conf.sizeRange.upper),
            separate=False)
        assert len (ssCleavage)   == numSites
        assert len (ssBgCleavage) == numSites

        # Save raw cut counts.
        for cleavage, outFileName in zip (
                (ssCleavage,              ssBgCleavage),
                (conf.outfile.rawends.ss, conf.outfile.rawends.ssbg)):
            if outFileName is not None:
                outFile = open (outFileName.replace ('%', feat.name), 'w')
                print >> outFile, '\n'.join ([str (int (x)) for x in cleavage])
        
        if doRawWiggle:
            # Write the wiggle stanza for this feature.
            # We use the convention that a value at position N describes the
            # between-base "site" immediately DOWNstream of it in the feature
            # (regardless of strand).
            # The 5' and 3' ends of the feature are not plotted.
            # So, there are N-1 such sites in a feature of length N for which
            # we plot values (that is why we subset "cleavage" arrays to
            # size N-1).
            # Therefore, you will never see a value for the last (3'-most) nuc
            # of a feature -- as that would describe the 3'-most end/site,
            # which we are ignoring.

            if feat.strand == '+':
                wiggleOutSs   = wiggleOutRawSsPlus
                wiggleOutSsBg = wiggleOutRawSsBgPlus
            else:
                assert feat.strand == '-'
                wiggleOutSs   = wiggleOutRawSsMinus
                wiggleOutSsBg = wiggleOutRawSsBgMinus

            # The starting position of between-base sites.
            if feat.strand == '+':                
                siteStart = feat.L
            else:
                siteStart = feat.L + 1

            wiggleOutSs.writeStanza (
                'fixedStep',
                chrom      = feat.landmark,
                start      = siteStart,
                strand     = feat.strand,
                step       = 1,
                span       = 1,
                values     = ssCleavage[1:-1],
                threshold  = 0,  # don't display 0 counts
                )

            wiggleOutSsBg.writeStanza (
                'fixedStep',
                chrom      = feat.landmark,
                start      = siteStart,
                strand     = feat.strand,
                step       = 1,
                span       = 1,
                values     = ssBgCleavage[1:-1],
                threshold  = 0,  # don't display 0 counts
                )

        # Find sites where we have no data in either experiment or
        # background sample.
        # We consider as "missing/no data" any site where counts are
        # sufficiently close to 0 as to likely be noise.
        noDataSs = set()
        for i in xrange (numSites):
            if ssCleavage[i] < conf.noiseCutoff \
                    and ssBgCleavage[i] < conf.noiseCutoff:
                noDataSs.add (i)
                ssCleavage[i]   = 0   # Wipe so that norm. constant and other
                ssBgCleavage[i] = 0   # calculations later are not affected.

        # Trim off sites at the end to prevent them affecting normalization
        # and so forth?
        if conf.numEndSitesToIgnore > 0:
            for offset in xrange (0, min (conf.numEndSitesToIgnore, numSites)):
                noDataSs.add (offset)
                noDataSs.add (numSites - 1 - offset)
                for sample in (ssCleavage, ssBgCleavage):
                    # Wipe, so that normalization constant and other
                    # calculations later are not affected.
                    sample[offset]                = 0
                    sample[numSites - 1 - offset] = 0
        
        print >> statsFile, "Missing data sites in SS (0-based): ", \
            ', '.join (map (str, sorted (noDataSs)))

        # Add pseudocounts.
        # If 'ignoreNoDataSites' option is on, we only add pseudocounts
        # to nucs that have data in either nuclease sample or background
        # sample or both.
        # This way pseudocounts on no data won't affect the normalizing
        # constant.
        
        if conf.pseudocounts.auto:
            # Automatially set pseudocounts using a flat prior, conditioned on
            # the number of end counts (~ coverage) for this RNA.
            # That is, given the number of breaks we observe, we assume the
            # breaks are equally distributed amongst all positions.
            # Note that we take square root of sum of breaks, to satisfy the
            # condition that real data must dominate pseudocounts in the limit
            # to infinity.
            print >> statsFile, "Auto-setting pseudocounts..."
            
            if sum (ssCleavage) == 0 or numSites == len (noDataSs):
                print >> statsFile, "WARNING: Dummy pseudocount for SS"
                pseudocountSs = 1.
            else:
                pseudocountSs = \
                    math.sqrt (sum (ssCleavage) / (numSites - len (noDataSs)))

            if sum (ssBgCleavage) == 0 or numSites == len (noDataSs):
                print >> statsFile, "WARNING: Dummy pseudocount for SSBG"
                pseudocountSsBg = 1.
            else:
                pseudocountSsBg = \
                    math.sqrt (sum (ssBgCleavage) / (numSites - len (noDataSs)))

        else:
            print >> statsFile, "Using pseudocounts from config file..."
            pseudocountSs   = conf.pseudocounts.ss
            pseudocountSsBg = conf.pseudocounts.ssbg

        print >> statsFile, "Pseudocounts: ss %f ssbg %f" \
              % (pseudocountSs, pseudocountSsBg)
        
        for i in xrange (numSites):
            if not conf.ignoreNoDataSites or i not in noDataSs:
                ssCleavage[i]   += pseudocountSs
                ssBgCleavage[i] += pseudocountSsBg
        
        for cleavage, sampleName in zip (
                (ssCleavage, ssBgCleavage),
                ('ss',       'ssbg')):
            print >> statsFile, "Cut counts (with pseudocounts) in %s: %s" % \
                (sampleName, ', '.join (map (str, cleavage)))
        
        # Normalize num cuts per site so that sum of total cuts per sample
        # is the same across all samples.
        for sample, sampleName in zip (
                (ssCleavage, ssBgCleavage),
                ('ss',       'ssbg')):
            # Normalize.
            numCutsTotal = sum (sample)
            print >> statsFile, "Normalizing constant (%s): %d" % \
                (sampleName, numCutsTotal)
            # Normalizing constant can be 0 in cases of no data
            # for any nuc, so have to check.
            if numCutsTotal > 0:
                for i in xrange (len (sample)):
                    sample[i] /= float (numCutsTotal)

                # Safety, allowing for imprecision.
                s = sum (sample)
                assert s > 0.99 and s < 1.01

        for cleavage, sampleName in zip (
                (ssCleavage, ssBgCleavage),
                ('ss',       'ssbg')):
            print >> statsFile, "Normalized cut counts in %s: %s" % \
                (sampleName, ', '.join (map (str, cleavage)))

        if doProbWiggle:
            # Write the wiggle stanza for this feature, showing break probabilities.
            # We use the convention that a value at position N describes the
            # between-base "site" immediately DOWNstream of it in the feature
            # (regardless of strand).
            # The 5' and 3' ends of the feature are not plotted.
            # So, there are N-1 such sites in a feature of length N for which
            # we plot values (that is why we subset "cleavage" arrays to
            # size N-1).
            # Therefore, you will never see a value for the last (3'-most) nuc
            # of a feature -- as that would describe the 3'-most end/site,
            # which we are ignoring.

            if feat.strand == '+':
                wiggleOutSs   = wiggleOutProbSsPlus
                wiggleOutSsBg = wiggleOutProbSsBgPlus
            else:
                assert feat.strand == '-'
                wiggleOutSs   = wiggleOutProbSsMinus
                wiggleOutSsBg = wiggleOutProbSsBgMinus

            # The starting position of between-base sites.
            if feat.strand == '+':                
                siteStart = feat.L
            else:
                siteStart = feat.L + 1

            wiggleOutSs.writeStanza (
                'fixedStep',
                chrom      = feat.landmark,
                start      = siteStart,
                strand     = feat.strand,
                step       = 1,
                span       = 1,
                values     = ssCleavage[1:-1],
                threshold  = 0,  # don't display 0 counts
                )

            wiggleOutSsBg.writeStanza (
                'fixedStep',
                chrom      = feat.landmark,
                start      = siteStart,
                strand     = feat.strand,
                step       = 1,
                span       = 1,
                values     = ssBgCleavage[1:-1],
                threshold  = 0,  # don't display 0 counts
                )

        # Correct counts for background in-place.
        correctForBg (
            ssCleavage, ssBgCleavage, useLog=conf.logScores)

        if conf.ignoreNoDataSites:
            # Some outputs require None type to denote missing data.
            # Convert array to list and put it there.
            ssCleavage = [
                None if i in noDataSs else ssCleavage[i]
                for i in xrange (len (ssCleavage))]

        if conf.posOnly:
            # Blank out all zero and negative scores by replacing them
            # with "missing data".
            for i in xrange (numSites):
                if ssCleavage[i] is not None and ssCleavage[i] <= 0.0:
                    ssCleavage[i] = None
                    noDataSs.add (i)

        if doCutScoreWiggle:
            # Write the wiggle stanza for this feature.
            # We use the convention that a value at position N describes the
            # between-base "site" immediately DOWNstream of it in the feature
            # (regardless of strand).
            # The 5' and 3' ends of the feature are not plotted.
            # So, there are N-1 such sites in a feature of length N for which
            # we plot values (that is why we subset "cleavage" arrays to
            # size N-1).
            # Therefore, you will never see a value for the last (3'-most) nuc
            # of a feature -- as that would describe the 3'-most end/site,
            # which we are ignoring.

            if feat.strand == '+':
                wiggleOutSs = wiggleOutCutScoreSsPlus
            else:
                assert feat.strand == '-'
                wiggleOutSs = wiggleOutCutScoreSsMinus

            # The starting position of between-base sites.
            if feat.strand == '+':                
                siteStart = feat.L 
            else:
                siteStart = feat.L + 1

            logger.debug (
                'Writing SS wiggle data for "%s" to file...' % feat.name)
            wiggleOutSs.writeStanza (
                'fixedStep',
                chrom      = feat.landmark,
                start      = siteStart,
                strand     = feat.strand,
                step       = 1,
                span       = 1,
                values     = ssCleavage[1:-1],
                )

        # Output cut scores/rates per site.
        if conf.outfile.cutscores.ss is not None:
            outFileSs = open (
                conf.outfile.cutscores.ss.replace ('%', feat.name), 'w')
            for i in xrange (len (ssCleavage)):
                if conf.ignoreNoDataSites and i in noDataSs:
                    print >> outFileSs, 'none'
                else:
                    print >> outFileSs, str (ssCleavage[i])

